import datetime
import json
import os
import random
import re
import shutil
import string
import sys
import time

import click
import netifaces

import nbeam.server
from nbeam.api import request_ssl, get_ssl, renew_ssl, get_renewal

HOME = os.environ.get('HOME', os.path.abspath(os.getcwd()))
DEFAULT_CONFIG = os.path.join(HOME, '.config', 'neutron-beam-64', 'config.json')
SETUP_MESSAGE = 'Configuration not found please run "nbeam setup_remote" or "nbeam setup_local" first.'


def get_config(context):
  if os.path.exists(context.obj['config']):
    return parse_config(context)

  else:
    click.echo(SETUP_MESSAGE)


def parse_config(context):
  with open(context.obj['config'], 'r') as fh:
    config = json.load(fh)

  config['config'] = context.obj['config']
  config['debug'] = context.obj['debug']
  return config


@click.group()
@click.option('--config', default=DEFAULT_CONFIG, type=click.Path(exists=False))
@click.option('--debug/--no-debug', default=False)
@click.pass_context
def cli(context, config, debug):
  context.obj['config'] = os.path.abspath(config)
  context.obj['debug'] = debug


def update_cert(context,
                config,
                chain=None,
                privkey=None,
                created=None,
                expires=None):
  if chain is None or privkey is None or created is None:
    return

  config['created'] = created

  with open(context.obj['config'], 'w') as fh:
    fh.write(json.dumps(config, indent=2))

  ssl_opts = config['ssl_options']
  domain = config['domain']

  with open(ssl_opts['certfile'], 'w') as cert:
    cert.write(chain)

  with open(ssl_opts['keyfile'], 'w') as key:
    key.write(privkey)

  print('Certificate updated.')
  print('Expires:', expires)


def write_certs(context, chain, pkey):
  config_dir = os.path.dirname(context.obj['config'])
  if not os.path.exists(config_dir):
    os.makedirs(config_dir)

  chain_path = os.path.join(config_dir, 'fullchain.pem')
  pkey_path = os.path.join(config_dir, 'privkey.pem')

  with open(chain_path, 'w') as fh:
    fh.write(chain)

  with open(pkey_path, 'w') as fh:
    fh.write(pkey)

  return chain_path, pkey_path


def write_config(context,
                 key,
                 cert,
                 address,
                 terminal,
                 workspace,
                 user,
                 domain=None):
  api_key = ''.join([
      random.SystemRandom().choice("{}{}{}".format(
          string.ascii_letters, string.digits, string.punctuation))
      for i in range(50)
  ])

  created = datetime.datetime.now(datetime.timezone.utc).isoformat()

  config = {
      'port': 32789,
      'address': address,
      'ssl_options': {
          'keyfile': key,
          'certfile': cert,
      },
      'user': user,
      'workspaces': [workspace],
      'api_key': api_key,
      'domain': domain,
      'home': HOME,
      'created': created,
      "terminal": {
          "enabled": terminal,
          "shell": shutil.which("bash"),
          "timeout": 60 * 5
      },
  }

  config_dir = os.path.dirname(context.obj['config'])
  if not os.path.exists(config_dir):
    os.makedirs(config_dir)

  with open(context.obj['config'], 'w') as fh:
    fh.write(json.dumps(config, indent=2))

  click.echo('{} written.'.format(context.obj['config']))
  click.echo('API Key: {}'.format(api_key))


def validate_user(ctx, param, value):
  while True:
    if re.search('^User\d+$', value):
      return value

    click.echo('User must match format "UserXXX"')
    value = click.prompt(param.prompt)


def ssl_flow(context, ip=None):
  url = request_ssl(ip)
  click.echo('Go to the following URL to finish your request:')
  click.echo(url)
  click.echo('\n')
  vcode = click.prompt('Verification code')

  if vcode == 'skip':
    return os.path.join(HOME, 'cert'), os.path.join(HOME,
                                                    'key'), 'skip.example.com'

  data = get_ssl(url, vcode)
  chain, pkey = write_certs(context, data['chain'], data['privkey'])
  return chain, pkey, data['domain']


@cli.command()
@click.pass_context
@click.option(
    '--key', type=click.Path(exists=True, dir_okay=False), default=None)
@click.option(
    '--cert', type=click.Path(exists=True, dir_okay=False), default=None)
@click.option(
    '--terminal/--no-terminal', prompt='Terminal Enabled', default=True)
@click.option(
    '--workspace',
    default=HOME,
    prompt='Workspace directory',
    type=click.Path(exists=True, file_okay=False))
@click.option('--user', prompt='User ID', callback=validate_user)
def setup_remote(context, key, cert, terminal, workspace, user):
  domain = None
  if key and cert:
    key = os.path.abspath(key)
    cert = os.path.abspath(cert)

  else:
    cert, key, domain = ssl_flow(context, None)

  write_config(context, key, cert, '', terminal, workspace, user, domain=domain)
  if domain is not None:
    click.echo('Address: {}'.format(domain))


@cli.command()
@click.pass_context
@click.option(
    '--terminal/--no-terminal', prompt='Terminal Enabled', default=True)
@click.option(
    '--workspace',
    default=HOME,
    prompt='Workspace directory',
    type=click.Path(exists=True, file_okay=False))
@click.option('--user', prompt='User ID', callback=validate_user)
def setup_local(context, terminal, workspace, user):
  address = '127.0.0.1'
  write_config(context, None, None, address, terminal, workspace, user)


@cli.command()
@click.pass_context
@click.option('--ip', default=None)
@click.option(
    '--key', type=click.Path(exists=True, dir_okay=False), default=None)
@click.option(
    '--cert', type=click.Path(exists=True, dir_okay=False), default=None)
@click.option(
    '--terminal/--no-terminal', prompt='Terminal Enabled', default=True)
@click.option(
    '--workspace',
    default=HOME,
    prompt='Workspace directory',
    type=click.Path(exists=True, file_okay=False))
@click.option('--user', prompt='User ID', callback=validate_user)
def setup_termux(context, ip, key, cert, terminal, workspace, user):
  domain = None
  if key and cert:
    key = os.path.abspath(key)
    cert = os.path.abspath(cert)

  else:
    if ip is None:
      ifaces = netifaces.interfaces()
      if 'arc0' in ifaces:
        ip = netifaces.ifaddresses('arc0')[netifaces.AF_INET][0]['addr']
      elif 'eth0' in ifaces:
        ip = netifaces.ifaddresses('eth0')[netifaces.AF_INET][0]['addr']
      else:
        raise Exception('Could not find your IP address, use the --ip option')

    cert, key, domain = ssl_flow(context, ip)

  write_config(context, key, cert, '', terminal, workspace, user, domain=domain)
  if domain is not None:
    click.echo('Address: {}'.format(domain))


@cli.command()
@click.pass_context
def apikey(context):
  config = get_config(context)
  if config:
    click.echo('API Key: {}'.format(config['api_key']))

  else:
    click.echo('No Config Found')


@cli.command()
@click.pass_context
def domain(context):
  config = get_config(context)
  if config:
    click.echo('Domain: {}'.format(config['domain']))

  else:
    click.echo('No Config Found')


@cli.command()
@click.pass_context
@click.option(
    '--force/--no-force', default=False, help='Force start if pid file exists.')
def start(context, force):
  config = get_config(context)
  if config:
    click.echo('Started Neutron Beam')
    nbeam.server.start(force=force, **config)


@cli.command()
@click.pass_context
def stop(context):
  config = get_config(context)
  if config:
    click.echo('Stopping Neutron Beam ...', nl=False)
    nbeam.server.stop(**config)
    click.echo(' Stopped')


@cli.command()
@click.pass_context
def restart(context):
  context.forward(stop)
  context.forward(start)


@cli.command()
@click.pass_context
def renew(context):
  config = get_config(context)
  if 'ssl_options' not in config:
    print('Configuration missing SSL Options')
    return

  if 'domain' not in config:
    print('Configuration missing Domain')
    return

  ssl_opts = config['ssl_options']
  domain = config['domain']

  with open(ssl_opts['certfile'], 'r') as cert:
    with open(ssl_opts['keyfile'], 'r') as key:
      response = renew_ssl(domain, cert.read(), key.read())

  if response is not None:
    vcode = response['vcode']
    print('Waiting for certificate to renew ...')
    time.sleep(60)
    count = 0

    while 1:
      time.sleep(15)
      response = get_renewal(domain, vcode)
      if response is not None and 'chain' in response:
        update_cert(context, config, **response)
        sys.exit(0)

      count += 1
      if count > 10:
        print('Certificate renewal failed.')
        sys.exit(1)

      print('Trying again ....')

  else:
    print('Certificate renewal failed.')
    sys.exit(1)


def main():
  cli(obj={})


if __name__ == '__main__':
  cli(obj={})
