import os
import json

import jwt
from terminado import TermSocket

import tornado.httpserver
import tornado.log
import tornado.ioloop
import tornado.options
import tornado.web
import tornado.websocket

import nbeam
from nbeam.auth import request_is_valid, jwt_is_valid
from nbeam.utils import valid_path
from nbeam.settings import ALLOWED_ORIGINS
import nbeam.tasks
from nbeam.tmanager import TerminalManager

if os.name == 'nt':
  daemon = None

else:
  import nbeam.daemon as daemon


class TerminalSocket(TermSocket):

  def check_origin(self, origin):
    if origin in ALLOWED_ORIGINS:
      return True

    return False

  def open(self, *args, **kwargs):
    jwt = self.get_argument('jwt')
    if jwt_is_valid(jwt, self.settings):
      return super().open(*args, **kwargs)

    self.close()


class BeamWebSocket(tornado.websocket.WebSocketHandler):

  def check_origin(self, origin):
    if origin in ALLOWED_ORIGINS:
      return True

    return False

  def write_json(self, data, rid):
    payload = {'user': self.settings['user']}
    encoded = jwt.encode(payload, self.settings['api_key'], algorithm='HS256')

    data['id'] = rid
    data['jwt'] = encoded.decode("utf-8")

    self.write_message(data)

  def error(self, error, rid):
    self.write_json({'error': error}, rid)

  def route(self, data):
    tornado.log.logging.info('Task: ' + data['task'])
    task = getattr(nbeam.tasks, 'nbeam_' + data['task'])

    return task(data, self.settings)

  def open(self):
    tornado.log.logging.info("WebSocket opened")

  def on_message(self, message):
    data = json.loads(message)

    if request_is_valid(data, self.settings):
      response = self.route(data)
      self.write_json(response, data['id'])

    else:
      self.error('Request is not valid.', data['id'])

  def on_close(self):
    tornado.log.logging.info("WebSocket closed")


class VersionHandler(tornado.web.RequestHandler):

  def get(self):
    self.set_header("Content-Type", "text/plain")
    self.write("Version: " + nbeam.version)


class StaticFileHandler(tornado.web.StaticFileHandler):

  def initialize(self, default_filename=None):
    self.default_filename = default_filename

  @property
  def root(self):
    return self.settings['workspaces'][self.request.workspace]

  def get(self, workspace, token, path, **kwargs):
    self.request.workspace = int(workspace) - 1

    if jwt_is_valid(token, self.settings):
      if valid_path(path, {'workspace': workspace}, self.settings):
        return super().get(path, **kwargs)

    self.set_status(401)
    self.write('Not Authorized')


def make_app(settings):
  handlers = [
      (r"/view/(\d+)/(\S+?)/(.*)", StaticFileHandler),
      (r"/beam", BeamWebSocket),
      (r"/version", VersionHandler),
      (r"/", VersionHandler),
  ]

  term_manager = None
  if settings['terminal']['enabled']:
    shell = settings['terminal']['shell']
    term_manager = TerminalManager(shell_command=[shell])
    handlers.insert(0, (r"/terminal/(\S+)", TerminalSocket, {
        'term_manager': term_manager
    }))

  return tornado.web.Application(handlers, **settings), term_manager


def start_tornado(**settings):
  if not settings['debug']:
    tornado.options.options.log_file_prefix = log_file(**settings)

  tornado.log.enable_pretty_logging()

  ssl_options = None
  if settings['ssl_options']['keyfile'] and settings['ssl_options']['certfile']:
    ssl_options = settings['ssl_options']

  port = settings['port']
  address = settings['address']

  if settings['home']:
    os.chdir(settings['home'])

  del settings['ssl_options']
  del settings['port']
  del settings['address']
  del settings['home']

  app, term_manager = make_app(settings)

  http_server = tornado.httpserver.HTTPServer(app, ssl_options=ssl_options)
  http_server.listen(port, address=address)

  loop = tornado.ioloop.IOLoop.current()

  try:
    loop.start()

  finally:
    if term_manager:
      term_manager.shutdown()

    loop.close()


def generate_runner(**settings):

  class D(daemon.Daemon):

    def run(self):
      start_tornado(**settings)

  return D


def log_file(**settings):
  d = os.path.dirname(settings['config'])
  return os.path.join(d, 'nbeam.log')


def pid_file(**settings):
  d = os.path.dirname(settings['config'])
  return os.path.join(d, 'nbeam.pid')


def start(**settings):
  if daemon and not settings['debug']:
    runnerClass = generate_runner(**settings)
    pid = pid_file(**settings)
    if settings['force']:
      if os.path.exists(pid):
        os.remove(pid)

    runner = runnerClass(pid, verbose=0)
    runner.start()

  else:
    start_tornado(**settings)


def stop(**settings):
  if daemon and not settings['debug']:
    runnerClass = generate_runner(**settings)
    runner = runnerClass(pid_file(**settings), verbose=0)
    runner.stop()
